package cz.upol.pja.lecture10.examples;

import static cz.upol.pja.lecture10.jbc.Util.typeSignature;

import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;

import cz.upol.pja.lecture10.jbc.ClassBuilder;
import cz.upol.pja.lecture10.jbc.MethodBuilder;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.Bytecode;
import javassist.bytecode.Opcode;

public class MethodCalls {
	
	private static final String WORKING_DIRECTORY = System.getProperty("java.io.tmpdir");
	private static final String CLASS_NAME = "ThirdClass";
	
	public static void main(String[] args) throws Exception {
		ClassBuilder thirdClassBuilder = new ClassBuilder(CLASS_NAME);
		
		createHelloWorldMethod(thirdClassBuilder);
		createFactRecursive(thirdClassBuilder);
		createPointMethod(thirdClassBuilder);
		
		thirdClassBuilder.save(WORKING_DIRECTORY);
		
		testMethods();
	}
	
	/**
	 * Vytvori metodu odpovidajici deklaraci
	 * public static void helloWorld() {
	 *    System.out.println("Hello World!");
	 * } 
	 */
	private static void createHelloWorldMethod(ClassBuilder classBuilder) throws NotFoundException {
		MethodBuilder helloMethod = classBuilder.createMethod("helloWorld", Modifier.PUBLIC | Modifier.STATIC, CtClass.voidType);
		
		Bytecode bc = helloMethod.createByteCode(3, 1);
		
		// vlozime do poolu constant retezec, na ktery se budeme odkazovat
		int sHello = helloMethod.getConstantPool().addStringInfo("Hello World!");
		
		final ClassPool classPool = ClassPool.getDefault();
		
		// reprezentace jednotlivych trid, se kterymi budeme pracovat
		final CtClass classJavaLangSystem = classPool.get("java.lang.System");
        final CtClass classJavaIoPrintStream = classPool.get("java.io.PrintStream");
        final CtClass classString = classPool.get("java.lang.String");
        
        // samotne generovani kodu
        // ziskame hodnotu statickeho atributu "out" z tridy java.lang.System, ktery je typu java.io.PrintStream
        bc.addGetstatic(classJavaLangSystem, "out", typeSignature(classJavaIoPrintStream));

        // nacteme na zasobnik odkaz na konstantu
        bc.addLdc(sHello);
        
        // zavolame metodu java.io.PrintStream.println
        bc.addInvokevirtual(classJavaIoPrintStream, "println", CtClass.voidType, new CtClass[] { classString });		
		
		bc.addOpcode(Opcode.RETURN);		
	}
	
	/**
	 * Vytvori metodu odpovidajici deklaraci:
	 * public static void factRec(int n) {
	 *   if (n > 0) return n * factRec(n - 1);
	 *   return 1;
	 * } 
	 */
	private static void createFactRecursive(ClassBuilder classBuilder) {
		MethodBuilder factMethod = classBuilder.createMethod("factRec", Modifier.PUBLIC | Modifier.STATIC, CtClass.intType, CtClass.intType);
		Bytecode bc = factMethod.createByteCode(3, 1);
		
		// if (n > 0) goto ...
		bc.addIload(0);
		int labelOperandIndex = bc.currentPc();
		bc.addOpcode(Opcode.IFGT);
		bc.addGap(2);
		
		// return 1
		bc.addIconst(1);
		bc.addOpcode(Opcode.IRETURN);
		
		// PATCH
		int offset = bc.currentPc() - labelOperandIndex;
		bc.write(labelOperandIndex + 1, offset >> 8);
		bc.write(labelOperandIndex + 2, offset);

		bc.addIload(0);

		// kopie n
		bc.addOpcode(Opcode.DUP);
		
		// n - 1
		bc.addIconst(1);
		bc.add(Opcode.ISUB);
		
		// rekurzivni volani
		bc.addInvokestatic(factMethod.getGeneratedClass(), "factRec", CtClass.intType, new CtClass[] { CtClass.intType });
		
		// nasobeni
		bc.add(Opcode.IMUL);

		bc.addOpcode(Opcode.IRETURN);
	}
	
	/**
	 * Vytvori metodu odpovidajici deklaraci:
	 * public static java.awt.Point createPoint() {
	 *    return new Point(42, 43);
	 * }
	 */
	public static void createPointMethod(ClassBuilder classBuilder) throws NotFoundException {
		CtClass classPoint = ClassPool.getDefault().get("java.awt.Point");
		
		MethodBuilder factMethod = classBuilder.createMethod("createPoint", Modifier.PUBLIC | Modifier.STATIC, classPoint);
		Bytecode bc = factMethod.createByteCode(5, 0);
		
		
		bc.addNew(classPoint);
		bc.add(Opcode.DUP);
		
		bc.addIconst(74);
		bc.addIconst(75);
		bc.addInvokespecial(classPoint, "<init>", CtClass.voidType, new CtClass[] { CtClass.intType, CtClass.intType });
		bc.add(Opcode.ARETURN);
	}
	
	/**
	 * Otestuje metody vytvorene v metode {@link #main}
	 */
	public static void testMethods() throws Exception {
		URL url = new URL("file:" + WORKING_DIRECTORY + "/");
		URLClassLoader cl = new URLClassLoader(new URL[] { url });

		Class<?> c = cl.loadClass(CLASS_NAME);
		
		Method m1 = c.getMethod("helloWorld");
		m1.invoke(null);
		
		Method m2 = c.getMethod("factRec", int.class);
		Object x2 = m2.invoke(null, 5);
		System.out.println(x2);
		
		Method m3 = c.getMethod("createPoint");
		Object x3 = m3.invoke(null);
		System.out.println(x3);	
		
		cl.close();
	}
}